#!/usr/bin/env python3
#################################################################
# @Program: miRvestigator.py                                    #
# @Version: 1                                                   #
# @Author: Chris Plaisier                                       #
# @Sponsored by:                                                #
# Nitin Baliga, ISB                                             #
# Institute for Systems Biology                                 #
# 1441 North 34th Street                                        #
# Seattle, Washington  98103-8904                               #
# (216) 732-2139                                                #
# @Also Sponsored by:                                           #
# Luxembourg Systems Biology Grant                              #
#                                                               #
# If this program is used in your analysis please mention who   #
# built it. Thanks. :-)                                         #
#                                                               #
# Copyrighted by Chris Plaisier  8/12/2009                      #
#################################################################

from copy import deepcopy
from sys import stdout
import os, pickle
import argparse
import time
import json
import multiprocessing

from firm import pssm

# A class designed to compute and hold the information from analyzing
# miRNA seeds against motifs from 3' UTRs.
#
# Variables:
# miRNAver - version of miRBase.org used.
# miRNAs - list of miRNAs where each entry is unique, and names are appened for overlapping seeds.
# permKMers - list of possible Kmers given the seed length.
#
#
# Functions:
# addSorted(curList,newItem) - adds an entry into miRNAScores[pssm] so that the list is sorted in the end. Ties are possible and not handled at this level.
# allKmers(length) - creates a list of all possible Kmers given a specific length and alphabet.
# setMiRNAs(seedStart,seedEnd) - gets miRNAs from miRBase.org from the latest release.
# getTopHit(pssm.name) - gets top hit for a PSSM.
# getScoreList(pssm.name) - returns scores for all miRNA seeds for given PSSM, sorted of course.
#
class miRvestigator:
    # Initialize and start the run
    def __init__(self, pssms, seqs3pUTR, seedModel=[6,7,8], minor=True,
                 p5=True, p3=True, textOut=True,
                 wobble=True, wobbleCut=0.25, baseDir='', outName='',
                 use_multiprocessing=True):
        print('\nmiRvestigator analysis started...')
        self.pssms = pssms
        self.miRNAs = self.setMiRNAs(0,8,minor,p5,p3)
        # Trim sequences down
        self.miRNAs_6mer_1 = self.trimSeqs(deepcopy(self.miRNAs),0,6)
        self.miRNAs_6mer_2 = self.trimSeqs(deepcopy(self.miRNAs),1,7)
        self.miRNAs_6mer_3 = self.trimSeqs(deepcopy(self.miRNAs),2,8)
        self.miRNAs_7mer_m8 = self.trimSeqs(deepcopy(self.miRNAs),1,8)
        self.miRNAs_7mer_a1 = self.trimSeqs(deepcopy(self.miRNAs),0,7)
        self.miRNAs_8mer = self.trimSeqs(deepcopy(self.miRNAs),0,8)
        p3utrSeqs = 'X'.join(seqs3pUTR)

        if not baseDir == '':
            dirName = baseDir+'/miRNA'
        else:
            dirName = 'miRNA'
        if not os.path.exists(dirName):
            os.mkdir(dirName)

        if 6 in seedModel:
            print('Screening out 6mers not present in 3\' UTRs...')
            if not os.path.exists(dirName+'/permKMers_6mers.pkl'):
                permKMers_6mer = self.allKmers(6)
                tmpKMers = []
                for i in permKMers_6mer:
                    if not p3utrSeqs.find(i)==-1:
                        tmpKMers.append(i)
                self.permKMers_6mer = tmpKMers
                pklFile = open(dirName+'/permKMers_6mers.pkl','wb')
                pickle.dump(self.permKMers_6mer,pklFile)
            else:
                pklFile = open(dirName+'/permKMers_6mers.pkl','rb')
                self.permKMers_6mer = pickle.load(pklFile)
            pklFile.close()

        if 7 in seedModel:
            print('Screening out 7mers not present in 3\' UTRs...')
            if not os.path.exists(dirName+'/permKMers_7mers.pkl'):
                permKMers_7mer = self.allKmers(7)
                tmpKMers = []
                for i in permKMers_7mer:
                    if not p3utrSeqs.find(i)==-1:
                        tmpKMers.append(i)
                self.permKMers_7mer = tmpKMers
                pklFile = open(dirName+'/permKMers_7mers.pkl','wb')
                pickle.dump(self.permKMers_7mer,pklFile)
            else:
                pklFile = open(dirName+'/permKMers_7mers.pkl','rb')
                self.permKMers_7mer = pickle.load(pklFile)
            pklFile.close()

        if 8 in seedModel:
            print('Screening out 8mers not present in 3\' UTRs...')
            if not os.path.exists(dirName+'/permKMers_8mers.pkl'):
                permKMers_8mer = self.allKmers(8)

                tmpKMers = []
                for i in permKMers_8mer:
                    if not p3utrSeqs.find(i)==-1:
                        tmpKMers.append(i)
                self.permKMers_8mer = tmpKMers
                pklFile = open(dirName+'/permKMers_8mers.pkl','wb')
                pickle.dump(self.permKMers_8mer,pklFile)
            else:
                pklFile = open(dirName+'/permKMers_8mers.pkl','rb')
                self.permKMers_8mer = pickle.load(pklFile)
            pklFile.close()
        print('Done.')
        miRNAScores = {}
        cur = 1
        # Building HMM Model
        outMe = []
        if use_multiprocessing:
            pool = multiprocessing.Pool()
            args = zip(pssms, [wobble] * len(pssms))
            results = pool.map(self.detect_pssm, args)
            for item in results:
                outMe.extend(item)
        else:
            for pssm in pssms:
                res = self.detect_pssm((pssm, wobble))
                outMe.extend(res)

        print('miRvestigator analysis completed.')
        outFile = open(dirName+'/scores'+str(outName)+'.csv','w')
        outFile.write('pssm,miRNAs,match_type')
        outFile.write('\n'+'\n'.join(outMe))
        outFile.close()


    def detect_pssm(self, arg):
        pssm, wobble = arg
        result = []
        print('\n' + pssm.name)
        print('Building HMM model for '+str(pssm.consensus_motif())+'...')
        #miRNAScores[pssm.name] = ['NA','NA']  ## WW: actually we don't use it
        # Then setup the HMM
        ## States ##
        ## and Starting Probabilities ##
        # NM1 = no match 1
        # NM2 = no match 2
        # PSSMi = PSSM at spot i
        maxPSSMi = len(pssm.matrix)
        states = ['NM1', 'NM2']
        sp = {'NM1': float(1)/float(maxPSSMi+1), 'NM2': 0}
        # Add the PSSM states
        for i in range(maxPSSMi):
            states += ['PSSM'+str(i)]
            sp['PSSM'+str(i)] = float(1)/float(maxPSSMi+1)
            if wobble==True:
                states += ['WOBBLE'+str(i)]
                sp['WOBBLE'+str(i)] = 0
        ## Transition probabilities
        tp = {}
        # NM1
        nm1_2_nm1 = 0.01
        tp['NM1'] = { 'NM1': nm1_2_nm1, 'NM2': 0 }
        leftOver1 = float(1-nm1_2_nm1)/float(maxPSSMi)
        for i in range(maxPSSMi):
            tp['NM1']['PSSM'+str(i)] = leftOver1
            if wobble==True:
                tp['NM1']['WOBBLE'+str(i)] = 0 # Don't start a seed with a wobble
        # NM2
        tp['NM2'] = { 'NM1': 0, 'NM2': 1 }
        for i in range(maxPSSMi):
            tp['NM2']['PSSM'+str(i)] = 0
            if wobble==True:
                tp['NM2']['WOBBLE'+str(i)] = 0
        # PSSMis
        for i in range(maxPSSMi):
            tp['PSSM'+str(i)] = { 'NM1': 0, 'NM2': 0.01 }
            if wobble==True:
                tp['WOBBLE'+str(i)] = { 'NM1': 0, 'NM2': 0.01 }
            if i==(maxPSSMi-1):
                tp['PSSM'+str(i)]['NM2'] = 1
                if wobble==True:
                    tp['WOBBLE'+str(i)]['NM2'] = 1
            for j in range(maxPSSMi):
                if j == i+1:
                    if wobble==True:
                        # Allow wobbly matches if T is >= wobbleCut
                        if float(pssm.matrix[i+1][2])>=float(wobbleCut) or float(pssm.matrix[i+1][3])>=float(wobbleCut):
                            tp['PSSM'+str(i)]['PSSM'+str(j)] = 0.80
                            tp['PSSM'+str(i)]['WOBBLE'+str(j)] = 0.19
                        # Otherwise don't allow wobbly matches
                        else:
                            tp['PSSM'+str(i)]['PSSM'+str(j)] = 0.99
                            tp['PSSM'+str(i)]['WOBBLE'+str(j)] = 0
                        tp['WOBBLE'+str(i)]['PSSM'+str(j)] = 1
                        tp['WOBBLE'+str(i)]['WOBBLE'+str(j)] = 0
                    else:
                        tp['PSSM'+str(i)]['PSSM'+str(j)] = 0.99
                else:
                    tp['PSSM'+str(i)]['PSSM'+str(j)] = 0
                    if wobble==True:
                        tp['PSSM'+str(i)]['WOBBLE'+str(j)] = 0
                        tp['WOBBLE'+str(i)]['PSSM'+str(j)] = 0
                        tp['WOBBLE'+str(i)]['WOBBLE'+str(j)] = 0
        ## Emission probabilities
        ep = {}
        # NM1
        ep['NM1'] = { 'A': 0.25, 'C': 0.25, 'G': 0.25, 'T': 0.25 }
        # S1 - None
        # S2 - None
        # NM2
        ep['NM2'] = { 'A': 0.25, 'C': 0.25, 'G': 0.25, 'T': 0.25 }
        # PSSMis
        for i in range(maxPSSMi):
            ep['PSSM'+str(i)] = { 'A': pssm.matrix[i][0],
                                  'C': pssm.matrix[i][1],
                                  'G': pssm.matrix[i][2],
                                  'T': pssm.matrix[i][3] }
            if wobble==True:
                # If motif has both G and U probability greater than wobblecut or random (0.25)
                if float(pssm.matrix[i][2])>=float(wobbleCut) and float(pssm.matrix[i][3])>=float(wobbleCut):
                    ep['WOBBLE'+str(i)] = { 'A': 0.5, 'C': 0.5, 'G': 0, 'T': 0 }
                # If motif has G greater than wobblecut or random (0.25)
                elif float(pssm.matrix[i][2])>=float(wobbleCut):
                    ep['WOBBLE'+str(i)] = { 'A': 1, 'C': 0, 'G': 0, 'T': 0 }
                # If motif has U greater than wobblecut or random (0.25)
                elif float(pssm.matrix[i][3])>=float(wobbleCut):
                    ep['WOBBLE'+str(i)] = { 'A': 0, 'C': 1, 'G': 0, 'T': 0 }
                # Otherwise be random (0.25 x 4)
                else:
                    ep['WOBBLE'+str(i)] = { 'A': 0.25, 'C': 0.25, 'G': 0.25, 'T': 0.25 }
        print('Done.')
        print('Starting miRNA detection for '+str(pssm.consensus_motif())+':')
        # First try for perfect 8mer
        #  a. Do the Viterbi for all miRNAs
        vitMirnas = []
        maxVitP = 0
        testVitWiki = []
        for miRNA in self.miRNAs_6mer_1:
            #forVit_8mer = self.forwardViterbi(list(self.miRNAs_8mer[miRNA]), states, sp, tp, ep)
            forVit_8mer = self.viterbi(list(self.miRNAs_8mer[miRNA]), states, sp, tp, ep)
            forVit_8mer = [0,0,forVit_8mer[0]]
            if float(forVit_8mer[2]) > float(maxVitP):
                maxVitP = float(forVit_8mer[2])
                vitMirnas = [miRNA]
            elif float(forVit_8mer[2])==float(maxVitP):
                vitMirnas.append(miRNA)
        print(maxVitP, vitMirnas)
        totPs_8mer = []
        vitPs_8mer = []
        hits_8mer = 0
        for kMer in self.permKMers_8mer:
            #permVit = self.forwardViterbi(list(kMer), states, sp, tp, ep)
            permVit = self.viterbi(list(kMer), states, sp, tp, ep)
            permVit = [0,0,permVit[0]]
            if float(permVit[2]) > float(maxVitP):
                hits_8mer = 2
                break
            elif float(permVit[2]) == float(maxVitP):
                hits_8mer += 1
                if hits_8mer > 1:
                    break
            totPs_8mer.append(permVit[0])
            vitPs_8mer.append(permVit[2])
        if hits_8mer<=1:
            print('8mer match!')
            #outFile.write('\n' + pssm.name+','+'_'.join(vitMirnas)+',8mer')
            result.append(pssm.name+','+'_'.join(vitMirnas)+',8mer')

        if hits_8mer>1:
            # Then try for perfect 7mer-m8
            vitMirnas = []
            maxVitP = 0
            for miRNA in self.miRNAs_6mer_1:
                #forVit_7mer_m8 = self.forwardViterbi(list(self.miRNAs_7mer_m8[miRNA]), states, sp, tp, ep)
                forVit_7mer_m8 = self.viterbi(list(self.miRNAs_7mer_m8[miRNA]), states, sp, tp, ep)
                forVit_7mer_m8 = [0,0,forVit_7mer_m8[0]]
                if float(forVit_7mer_m8[2]) > float(maxVitP):
                    maxVitP = float(forVit_7mer_m8[2])
                    vitMirnas = [miRNA]
                elif float(forVit_7mer_m8[2])==float(maxVitP):
                    vitMirnas.append(miRNA)
            print(maxVitP, vitMirnas)
            totPs_7mer = []
            vitPs_7mer = []
            hits_7mer_m8 = 0
            for kMer in self.permKMers_7mer:
                #permVit = self.forwardViterbi(list(kMer), states, sp, tp, ep)
                permVit = self.viterbi(list(kMer), states, sp, tp, ep)
                permVit = [0,0,permVit[0]]
                if float(permVit[2]) > float(maxVitP):
                    hits_7mer_m8 = 2
                    break
                elif float(permVit[2]) >= float(maxVitP):
                    hits_7mer_m8 += 1
                    if hits_7mer_m8 > 1:
                        break
                totPs_7mer.append(permVit[0])
                vitPs_7mer.append(permVit[2])
            if hits_7mer_m8<=1:
                print('7mer-m8 match!')
                #outFile.write('\n'+pssm.name+','+'_'.join(vitMirnas)+',7mer_m8')
                result.append(pssm.name+','+'_'.join(vitMirnas)+',7mer_m8')

            # Finally try for perfect 7mer-a1
            vitMirnas = []
            maxVitP = 0
            for miRNA in self.miRNAs_6mer_1:
                #forVit_7mer_a1 = self.forwardViterbi(list(self.miRNAs_7mer_a1[miRNA]), states, sp, tp, ep)
                forVit_7mer_a1 = self.viterbi(list(self.miRNAs_7mer_a1[miRNA]), states, sp, tp, ep)
                forVit_7mer_a1 = [0,0,forVit_7mer_a1[0]]
                if float(forVit_7mer_a1[2]) > float(maxVitP):
                    maxVitP = float(forVit_7mer_a1[2])
                    vitMirnas = [miRNA]
                elif float(forVit_7mer_a1[2])==float(maxVitP):
                    vitMirnas.append(miRNA)
            print(maxVitP, vitMirnas)
            totPs_7mer = []
            vitPs_7mer = []
            hits_7mer_a1 = 0
            for kMer in self.permKMers_7mer:
                #permVit = self.forwardViterbi(list(kMer), states, sp, tp, ep)
                permVit = self.viterbi(list(kMer), states, sp, tp, ep)
                permVit = [0,0,permVit[0]]
                if float(permVit[2]) > float(maxVitP):
                    hits_7mer_a1 = 2
                    break
                elif float(permVit[2]) >= float(maxVitP):
                    hits_7mer_a1 += 1
                    if hits_7mer_a1 > 1:
                        break
                totPs_7mer.append(permVit[0])
                vitPs_7mer.append(permVit[2])
            if hits_7mer_a1<=1:
                print('7mer-a1 match!')
                #outFile.write('\n'+pssm.name+','+'_'.join(vitMirnas)+',7mer_a1')
                result.append(pssm.name+','+'_'.join(vitMirnas)+',7mer_a1')
        if hits_8mer>1 and hits_7mer_m8>1 and hits_7mer_a1>1:
            print('No match!')
            result.append(pssm.name + ',NA,NA')
        return result

    def getScore(self, forVit, vitPs):
        return (float(forVit[0])/float(forVit[2]))*float(self.getPValue(forVit[2],vitPs))

    # Generates all possible sequences with letters to the length of depth.
    def allKmers(self,depth,letters=['A','C','G','T'],seqs=[''],curdepth=0):
        newseqs = []
        for seq in seqs:
            for letter in letters:
                newseqs.append(seq + letter)
        if depth > curdepth:
            return(self.allKmers(depth,letters,newseqs,curdepth + 1))
        else:
            return(seqs)

    # Get the miRNAs to compare against
    def setMiRNAs(self,seedStart,seedEnd, minor=True, p5=True, p3=True):
        if not os.path.exists('mature.fa.gz'):
            print('\nDownloading miRNA seeds from miRBase.org...')
            # Grab down the latest miRNA data from mirbase.org:
            #  ftp://mirbase.org/pub/mirbase/CURRENT/mature.fa.gz
            from ftplib import FTP
            ftp1 = FTP('mirbase.org')
            ftp1.login()
            ftp1.cwd('/pub/mirbase/CURRENT/')
            # Get the miRBase.org version number for reference.
            self.miRNAver = (ftp1.pwd().split('/'))[-1]
            outFile = open('mature.fa.gz','wb')
            ftp1.retrbinary('RETR mature.fa.gz',outFile.write)
            outFile.close()
            ftp1.quit()
            print('Done.')
        else:
            print('\nUsing already downloaded miRNA seeds.')

        # Read in miRNAs: miRNAs are labeled by the hsa-* names and grabbing 2-8bp
        ### Could merge these as they come in so that don't do redundant, and also so that the labels are together
        import gzip
        miRNAFile = gzip.open('mature.fa.gz','r')
        miRNAs = {}
        while 1:
            miRNALine = miRNAFile.readline().decode("utf-8")
            seqLine = miRNAFile.readline().decode("utf-8")
            if not miRNALine:
                break
            # Get the miRNA name
            curMiRNA = (miRNALine.lstrip('>').split(' '))[0]
            if (curMiRNA.split('-'))[0]=='hsa':
                if (minor==True or curMiRNA.find('*')==-1) and (p5==True or curMiRNA.find('-5p')==-1) and (p3==True or curMiRNA.find('-3p')==-1):
                    # Now grab out the 2-8bp and do reverse complement on it
                    miRNAs[curMiRNA] = self.reverseComplement((seqLine.strip())[seedStart:seedEnd])
        miRNAFile.close()

        # How many distinct kMers in miRNAs
        miRNAuniq = {}
        for miRNA in miRNAs:
            if not miRNAs[miRNA] in miRNAuniq:
                miRNAuniq[miRNAs[miRNA]] = [miRNA]
            else:
                miRNAuniq[miRNAs[miRNA]].append(miRNA)
        # Then merge them
        miRNAsMerged = {}
        for seed in miRNAuniq:
            tmpName = '_'.join(miRNAuniq[seed])
            miRNAsMerged[tmpName] = seed
        return miRNAsMerged

    def trimSeqs(self, miRNAs, start, stop):
        tmp = {}
        for miRNA in miRNAs:
            tmp[miRNA] = miRNAs[miRNA][start:stop]
        return tmp

    # Complement
    def complement(self,seq):
        complement = {'A':'T', 'T':'A', 'C':'G', 'G':'C', 'N':'N', 'U':'A'}
        complseq = [complement[base] for base in seq]
        return complseq

    # Reverse complement
    def reverseComplement(self,seq):
        seq = list(seq)
        seq.reverse()
        return ''.join(self.complement(seq))

    # Reverse complement
    def reverseMe(self,seq):
        seq = list(seq)
        seq.reverse()
        return ''.join(seq)

    # Modified from From wikipedia to do both forward calculation and viterbi
    def forwardViterbi(self, obs, states, start_p, trans_p, emit_p):
       T = {}
       for state in states:
           ##          prob.           V. path  V. prob.
           T[state] = (start_p[state], [state], start_p[state])
       for output in obs:
           U = {}
           for next_state in states:
               total = float(0)
               argmax = None
               valmax = float(0)
               for source_state in states:
                   (prob, v_path, v_prob) = T[source_state]
                   p = emit_p[source_state][output] * trans_p[source_state][next_state]
                   prob *= p
                   v_prob *= p
                   total += prob
                   if v_prob > valmax:
                       argmax = v_path + [next_state]
                       valmax = v_prob
               U[next_state] = (total, argmax, valmax)
           T = U
       ## apply sum/max to the final states:
       total = float(0)
       argmax = None
       valmax = float(0)
       for state in states:
           (prob, v_path, v_prob) = T[state]
           total += prob
           if v_prob > valmax:
               argmax = v_path
               valmax = v_prob
       return (total, argmax, valmax)

    def viterbi(self, obs, states, start_p, trans_p, emit_p):
        V = [{}]
        path = {}

        # Initialize base cases (t == 0)
        for y in states:
            V[0][y] = start_p[y] * emit_p[y][obs[0]]
            path[y] = [y]

        # Run Viterbi for t > 0
        for t in range(1,len(obs)):
            V.append({})
            newpath = {}

            for y in states:
                (prob, state) = max([(V[t-1][y0] * trans_p[y0][y] * emit_p[y][obs[t]], y0) for y0 in states])
                V[t][y] = prob
                newpath[y] = path[state] + [y]

            # Don't need to remember the old paths
            path = newpath

        (prob, state) = max([(V[len(obs) - 1][y], y) for y in states])
        return (prob, path[state])


    # Decide whether to add and add at correct position
    # Strucuture of elements:
    # [pssm, getConsensus(pssms[pssm]), miRNA, miRNAs[miRNA], forVit]
    def addSorted(self,all,new):
        inserted = 0
        if len(all)>0:
            for i in range(len(all)):
                if new['vitPValue']<all[i]['vitPValue']:
                    all.insert(i,new)
                    inserted = 1
                    break
            if inserted==0:
                all.append(new)
        else:
            all.append(new)
        return all

    # Get p-value from a probability and a distribution
    def getPValue(self,prob,dist):
        a = float(len([i for i in dist if i>=prob]))
        b = float(len(dist))
        if not a==0:
            return a/b
        else:
            return 0

    # Get score list for a PSSM
    def getScoreList(self,pssmName):
        return self.miRNAScores[pssmName]

    # Get top hit(s) for a PSSM.
    # Returns either one or more miRNAs based on whether a clear winner or a tie.
    def getTopHit(self,pssmName):
        scoreList = self.getScoreList(pssmName)
        if scoreList[0]['vitPValue']<scoreList[1]['vitPValue']:
            return [scoreList[0]]
        else:
            retMe = []
            i = 0
            while scoreList[0]['vitPValue']==scoreList[i]['vitPValue']:
                retMe.append(scoreList[i])
                i += 1
            return retMe

    # Get the scores for a PSSM.
    # Returns all miRNAs which conatin the miRNA name.
    def getmiRNAHit(self,pssmName,miRNAname):
        scoreList = self.getScoreList(pssmName)
        retMe = []
        for i in range(len(scoreList)):
            if not scoreList[i]['miRNA.name'].find(miRNAname)==-1:
                scoreList[i]['rank'] = i
                retMe.append(scoreList[i])
        return retMe

    # Structure of elements:
    # [pssm, getConsensus(pssms[pssm]), miRNA, miRNAs[miRNA], forVit]
    # <tr><td>miRNA Name</td><td>Alignment Start<sup>*</sup></td><td>Alignment Stop<sup>*</sup></td><td>Alignment Length</td><td>Motif (Length)</td><td>Alignment Type</td><td>Alignment</td><td>P(Alignment)</td></tr>
    def outHtml(self, outMe):
        writeMe = ''
        # miRNA name
        writeMe += '<tr align=\'center\' valign=\'center\'><td>'+str(outMe[2])+'</td><td>'+str(outMe[3])+'</td>'
        alignment = outMe[4][1] # Grab the alignment
        alignment.pop() # Get rid of the extra state which is added by the forwardViterbi function
        start = 1
        if alignment[0]=='NM1':
            for i in alignment:
                if i=='NM1':
                    start += 1
        # Alignment
        seedAlign = ''
        seed = list(outMe[3])
        motifAlign = ''
        motif = list(outMe[1])
        alignMe = alignment
        aligned = ''
        lenMatch = 0
        # First get them zero'd to the same point
        if start>1:
            for i in range(start-1):
                seedAlign += seed.pop(0)
                aligned += ' '
                motifAlign += '-'
                alignMe.pop(0)
        if not alignMe[0]=='PSSM0' and not alignMe[0]=='WOBBLE0':
            for i in range(int(alignMe[0][4])):
                seedAlign += '-'
                aligned += ' '
                motifAlign += motif.pop(0)
        # Then add the parts that align
        while 1:
            if len(alignMe)==0 or alignMe[0]=='NM2':
                break
            seedAlign += seed.pop(0)
            if alignMe[0][0]=='P':
                aligned += '|'
            elif alignMe[0][0]=='W':
                aligned += ':'
            lenMatch += 1
            motifAlign += motif.pop(0)
            alignMe.pop(0)
        # Then do the ending
        if len(alignMe)>0:
            for i in alignMe:
                seedAlign += seed[0]
                seed = seed[1:]
                aligned += ' '
                motifAlign += '-'
            alignMe = []
        if len(motif)>0 and len(alignMe)==0:
            for i in motif:
                seedAlign += '-'
                aligned += ' '
                motifAlign += i
        writeMe += '<td>'+str(start)+'</td><td>'+str(start+lenMatch-1)+'</td><td>'+str(lenMatch)+'</td><td><font face="Courier New"><pre>Seed  '+str(seedAlign)+'\n      '+str(aligned)+'\nMotif '+str(motifAlign)+'</pre></font></td>'
        # P(Alignment)
        writeMe += '<td>'+str(outMe[4][0])+'</td><td>'+str(float(len([i for i in self.totPs if i>=outMe[4][0]]))/float(len(self.totPs)))+'</td><td>'+str(outMe[4][2])+'</td><td>'+str(float(len([i for i in self.vitPs if i>=outMe[4][2]]))/float(len(self.vitPs)))+'</td></tr>'

        return writeMe

    # Decide whether to add and add at correct position
    # Strucuture of elements:
    # input = [pssm, getConsensus(pssms[pssm]), miRNA, miRNAs[miRNA], forVit], totPs, vitPs (Ps are permuted probabilities)
    # output = [miRNAname,miRNAseed,AlignStart,AlignStop,AlignLength,SeedAlign,Align,MotifAlign,P(Total),P-valueTotal,P(Viterbi),P-valueViterbi]
    def outData(self,outMe,totP,vitP,compModel,fullSeed):
        output = []
        # miRNA name
        output += [outMe[2],self.reverseComplement(fullSeed)]
        alignment = outMe[4][1] # Grab the alignment
        alignment.pop() # Get rid of the extra state which is added by the forwardViterbi function
        start = 1
        if alignment[0]=='NM1':
            for i in alignment:
                if i=='NM1':
                    start += 1
        # Alignment
        seedAlign = ''
        seed = list(self.reverseMe(outMe[3]))
        motifAlign = ''
        motif = list(outMe[1])
        alignMe = alignment
        aligned = ''
        lenMatch = 0
        # First get them zero'd to the same point
        if start>1:
            for i in range(start-1):
                seedAlign += seed.pop(0)
                aligned += '_'
                motifAlign += '-'
                alignMe.pop(0)
        if len(alignMe)>0 and not alignMe[0]=='PSSM0' and not alignMe[0]=='WOBBLE0':
            if alignMe[0][0]=='P':
                upTo = int(alignMe[0][4])
            elif alignMe[0][0]=='W':
                upTo = int(alignMe[0][6])
            for i in range(upTo):
                seedAlign += '-'
                aligned += '_'
                motifAlign += motif.pop(0)
        # Then add the parts that align
        while 1:
            if len(alignMe)==0 or alignMe[0]=='NM2':
                break
            seedAlign += seed.pop(0)
            if alignMe[0][0]=='P':
                aligned += '|'
            elif alignMe[0][0]=='W':
                aligned += ':'
            lenMatch += 1
            motifAlign += motif.pop(0)
            alignMe.pop(0)
        # Then do the ending
        if len(alignMe)>0:
            for i in alignMe:
                seedAlign += seed[0]
                seed = seed[1:]
                aligned += '_'
                motifAlign += '-'
            alignMe = []
        if len(motif)>0 and len(alignMe)==0:
            for i in motif:
                seedAlign += '-'
                aligned += '_'
                motifAlign += i
        output += [start,start+lenMatch-1,lenMatch,compModel,"'"+motifAlign,"'"+aligned,"'"+seedAlign]
        # P(Alignment)
        output += [outMe[4][0],totP,outMe[4][2],vitP]
        return [str(i) for i in output]


DESCRIPTION = "miRvestigator.py - standalone miRvestigator"

if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=DESCRIPTION)
    parser.add_argument('seqs', help='seqs.txt file')
    parser.add_argument('pssms', help='pssms file')
    args = parser.parse_args()

    if not os.path.exists(args.seqs):
        raise Exception('seqs.txt file does not exist')
    if not os.path.exists(args.pssms):
        raise Exception('pssms.json file does not exist')

    with open(args.seqs) as infile:
        seqs = infile.readlines()

    pssms = []
    with open(args.pssms) as infile:
        pssms_json = json.load(infile)
        for pssm_json in pssms_json:
            pssms.append(pssm.pssm(biclusterName=pssm_json['name'],
                                   nsites=pssm_json['nsites'],
                                   eValue=pssm_json['evalue'],
                                   pssm=pssm_json['matrix']))

    t1 = time.time()
    m2m = miRvestigator(pssms, seqs, seedModel=[6,7,8], minor=True, p5=True,
                        p3=True, wobble=False, wobbleCut=0.25,
                        use_multiprocessing=True)
    with open('m2m_standalone.pkl', 'wb') as outfile:
        pickle.dump(m2m, outfile)
    t2 = time.time()
    print("Completed miRvestigator in {:.2f} minutes".format((t2-t1)/60.))



import os
import re
from setuptools import setup

NAME = 'isb_firm'
PACKAGES = ['firm']
DESCRIPTION = 'Framework for inference  of regulation by miRNAs'
LICENSE = 'LGPL V3'
URI = 'https://github.com/baliga-lab/firm2'
AUTHOR = 'Baliga Lab, Institute for Systems Biology'
VERSION = '1.0.0'

KEYWORDS = ['miRvestigator', 'firm']

# See trove classifiers
# https://testpypi.python.org/pypi?%3Aaction=list_classifiers

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "Natural Language :: English",
    "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Software Development :: Libraries :: Python Modules"
    ]

PACKAGE_DATA = {
    'firm': ['common/*', 'FreqFiles/*',
             'TargetPredictionDatabases/PITA/*',
             'TargetPredictionDatabases/TargetScan/*']
}

if __name__ == '__main__':
    setup(name=NAME, description=DESCRIPTION,
          license=LICENSE,
          url=URI,
          version=VERSION,
          author=AUTHOR,
          author_email='wwu@systemsbiology.net',
          maintainer=AUTHOR,
          maintainer_email='wwu@systemsbiology.net',
          keywords=KEYWORDS,
          packages=PACKAGES,
          zip_safe=False,
          classifiers=CLASSIFIERS,
          include_package_data=True, package_data=PACKAGE_DATA,
          scripts=['bin/firm', 'bin/firm-convertminer'])



import argparse
from firm.pssm import pssm
import gzip, os, re, sys
from subprocess import *
from multiprocessing import Pool, cpu_count, Manager
from collections import defaultdict
import json


def mirnas_for(mirna, mirna_ids):
    return [mirnas for m_name, mirnas in mirna_ids.items()
            if mirna_name_matches(mirna, m_name)]


def mirna_name_matches(a, b):
    if a == b:
        return 1
    if len(a) < len(b):
        re1 = re.compile(a + '[a-z]$')
        if re1.match(b):
            return 1
    else:
        re1 = re.compile(b + '[a-z]$')
        if re1.match(a):
            return 1
    return 0


def check_weederlauncher():
    """Check whether weederlauncher exists on this system"""
    for d in os.environ["PATH"].split(os.pathsep):
        filepath = os.path.join(d, 'weederlauncher')
        if os.path.isfile(filepath) and os.access(filepath, os.X_OK):
            return True
    return False


# Run weeder and parse its output
# First weederTFBS -W 6 -e 1, then weederTFBS -W 8 -e 2, and finally adviser
def run_weeder(params):
    seqFile, tmpdir, freqfiles = params
    print(".", end="", file=sys.stderr, flush=True)
    weeder_pssms = []
    percTargets = 50
    revComp = False

    # First run weederTFBS for 6bp motifs
    weeder_args = ' ' + str(seqFile) + ' HS3P small T50 F%s' % freqfiles
    if revComp:
        weeder_args += ' -S'
    errOut = open(os.path.join(tmpdir, 'weeder', 'stderr.out'), 'w')
    weederProc = Popen("weederlauncher " + weeder_args, shell=True, stdout=PIPE, stderr=errOut)
    output = weederProc.communicate()

    # Now parse output from weeder
    PSSMs = []
    with open(str(seqFile)+'.wee','r') as output:
        outLines = [line for line in output.readlines() if line.strip()]
    hitBp = {}
    # Get top hit of 6bp look for "1)"
    while 1:
        outLine = outLines.pop(0)
        if not outLine.find('1) ') == -1:
            break
    hitBp[6] = outLine.strip().split(' ')[1:]

    # Scroll to where the 8bp reads wll be
    while 1:
        outLine = outLines.pop(0)
        if not outLine.find('Searching for motifs of length 8') == -1:
            break

    # Get top hit of 8bp look for "1)"
    while 1:
        outLine = outLines.pop(0)
        if not outLine.find('1) ') == -1:
            break
    hitBp[8] = outLine.strip().split(' ')[1:]

    # Scroll to where the 8bp reads wll be
    while 1:
        outLine = outLines.pop(0)
        if not outLine.find('Your sequences:') == -1:
            break

    # Get into the highest ranking motifs
    seqDict = {}
    while 1:
        outLine = outLines.pop(0)
        if not outLine.find('**** MY ADVICE ****') == -1:
            break
        splitUp = outLine.strip().split(' ')
        seqDict[splitUp[1]] = splitUp[3].lstrip('>')

    # Get into the highest ranking motifs
    while 1:
        outLine = outLines.pop(0)
        if not outLine.find('Interesting motifs (highest-ranking)') == -1:
            break
    while 1:
        name = seqFile.split('/')[-1].split('.')[0] +'_'+ outLines.pop(0).strip() # Get match
        if not name.find('(not highest-ranking)') == -1:
            break
        # Get redundant motifs
        outLines.pop(0)
        redMotifs = [i for i in outLines.pop(0).strip().split(' ') if not i=='-']
        outLines.pop(0)
        outLines.pop(0)
        line = outLines.pop(0)
        instances = []
        while line.find('Frequency Matrix') == -1:
            splitUp = [i for i in line.strip().split(' ') if i]
            instances.append({'gene':seqDict[splitUp[0]],
                              'strand':splitUp[1],
                              'site':splitUp[2],
                              'start':splitUp[3],
                              'match':splitUp[4].lstrip('(').rstrip(')') })
            line = outLines.pop(0)
        # Read in Frequency Matrix
        outLines.pop(0)
        outLines.pop(0)
        matrix = []
        col = outLines.pop(0)
        while col.find('======') == -1:
            nums = [i for i in col.strip().split('\t')[1].split(' ') if i]
            colSum = 0
            for i in nums:
                colSum += int(i.strip())
            matrix += [[ float(nums[0]) / float(colSum),
                         float(nums[1]) / float(colSum),
                         float(nums[2]) / float(colSum),
                         float(nums[3]) / float(colSum)]]
            col = outLines.pop(0)
        weeder_pssms.append(pssm(name=name,
                                 sites=instances,
                                 evalue=hitBp[len(matrix)][1],
                                 pssm=matrix,
                                 genes=redMotifs))
    return weeder_pssms

def phyper(q, m, n, k):
    # Get an array of values to run
    rProc = Popen('R --no-save --slave', shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    runMe = []
    for i in range(len(q)):
        runMe.append('phyper('+str(q[i])+','+str(m[i])+','+str(n[i])+','+str(k[i])+',lower.tail=F)')
    runMe = ('\n'.join(runMe)+'\n').encode()
    out = rProc.communicate(runMe)
    return [line.strip().split(' ')[1]
            for line in out[0].decode("utf-8").strip().split('\n') if line]


def cluster_hypergeo(params):
    global db, dataset_genes, total_targets, mirna_target_dict
    cluster, cluster_genes, dataset, dboutdir = params
    sys.stderr.write(".")
    sys.stderr.flush()
    outpath = os.path.join(dboutdir, '%s_%s.csv' % (dataset, str(cluster)))
    with open(outpath, 'w') as outfile:
        outfile.write('miRNA,Cluster.Targets,miRNA.Targets,Cluster.Genes,Total,P.Value\n')
        # k = overlap, N = potential target genes, n = miRNA targets, m = cluster genes
        # Take gene list and compute overlap with each miRNA
        allGenes = set(dataset_genes).intersection(set(total_targets))
        genes = set(cluster_genes).intersection(set(allGenes))
        writeMe = []
        keys1 = mirna_target_dict.keys()
        m1s = []
        q = []
        m = []
        n = []
        k = []
        for m1 in keys1:
            m1s.append(m1)
            miRNAGenes = set(mirna_target_dict[m1]).intersection(allGenes)
            q.append(len(set(miRNAGenes).intersection(genes)))
            m.append(len(miRNAGenes))
            n.append(len(allGenes) - len(miRNAGenes))
            k.append(len(genes))
        results = phyper(q, m, n, k)
        for i in range(len(m1s)):
            writeMe.append(str(m1s[i]) + ',' + str(q[i]) + ',' + str(m[i]) + ',' + str(n[i]) + ',' + str(k[i]) + ',' + str(results[i]))
        outfile.write('\n'.join(writeMe))


# Benjamini-Hochberg - takes a dictionary of { name: pValue, ... }
def benjamini_hochberg(dict1, tests, alpha=0.001):
    # create a list of names sorted by the p-values in the input dictionary
    sorted1 = [i[0] for i in sorted(list(dict1.items()), key=lambda t: t[1])]

    # Then control based on FDR (false discovery rate)
    res1 = []
    alpha = float(alpha)
    for i in range(len(sorted1)):
        if dict1[sorted1[i]] <= alpha*(float(i+1)/float(tests)):
            res1.append(sorted1[i])
        else:
            break
    return res1


def make_mirna_dict(mirna_path):
    """reads the mature.fa.gz and returns a mapping from miRNA names => miRNA ids.
    This was changed from the original FIRM so we can use updated mature.fa.gz
    files instead of a prefiltered, potentially outdated file"""
    mirna_ids = {}
    with gzip.open(mirna_path, 'r') as infile:
        while True:
            desc = infile.readline()
            if not desc:
                break
            infile.readline()  # skip the sequence line
            desc = desc.decode('utf-8')
            if desc.startswith('>hsa'):
                comps = desc.lstrip('>').split(' ')
                mirna_name = comps[0].lower()
                mirna_id = comps[1]

                if not mirna_name in mirna_ids:
                    mirna_ids[mirna_name] = mirna_id
                else:
                    raise Exception("already exists: '%s'" % mirna_name)

    return mirna_ids


def make_refseq2entrez(gene2refseq_path):
    # Read in gene2refseq mappings and make a dictionary
    with gzip.open(gene2refseq_path, 'r') as inFile:
        refSeq2entrez = {}
        while 1:
            line = inFile.readline()
            if not line:
                break
            line = line.decode("utf-8")

            # Only add those that have the correct NCBI organism ID
            splitUp = line.strip().split('\t')
            if int(splitUp[0]) == 9606:
                # Check that the nucleotide ID is not a '-' and that it has
                # genomic coordiantes assocaited with it
                if not splitUp[3] == '-':
                    tmp = splitUp[3].split('.')[0]
                    if not tmp in refSeq2entrez:
                        refSeq2entrez[tmp] = int(splitUp[1])

        return refSeq2entrez


def read_sequences(seq_path):
    # 2. Read in sequences
    with gzip.open(seq_path, 'r') as seqFile:
        seqLines = [line.decode("utf-8") for line in seqFile.readlines()]
        ids = [i.strip().split(',')[0].upper() for i in seqLines]
        sequences = [i.strip().split(',')[1] for i in seqLines]
        seqs = dict(zip(ids,sequences))
    return seqs


def prepare_weeder_input(seqs, refSeq2entrez, use_entrez, exp_dir, tmpdir):
    # For each cluster file in exp from Goodarzi et al.
    # Cluster files should have a header and be tab delimited to look like this:
    # Gene\tGroup\n
    # NM_000014\t52\n
    # <RefSeq_ID>\t<signature_id>\n
    # ...
    fasta_files = []
    files = os.listdir(exp_dir)
    weeder_fasta_dir = os.path.join(tmpdir, 'weeder', 'fasta')
    if not os.path.exists(weeder_fasta_dir):
        os.makedirs(weeder_fasta_dir)

    for file in files:
        # 3. Read in cluster file and convert to entrez ids
        with open(os.path.join(exp_dir, file), 'r') as inFile:
            dataset = file.strip().split('.')[0]
            inFile.readline()
            lines = inFile.readlines()
            clusters = defaultdict(list)
            for line in lines:
                gene, group = line.strip().split('\t')
                group = int(group)
                if use_entrez:
                    entrez = int(gene)
                    clusters[group].append(entrez)
                else:
                    if gene in refSeq2entrez:
                        clusters[group].append(refSeq2entrez[gene])

        # 5. Make a FASTA file & run weeder
        for cluster in clusters:
            # Get sequences
            cluster_seqs = {}
            for target in clusters[cluster]:
                if str(target) in seqs:
                    cluster_seqs[target] = seqs[str(target)]
                else:
                    print("Did not find seq for '%s' (cluster %d)" % (target, cluster),
                          file=sys.stderr)

            # Make FASTA file
            fname = "%d_%s.fasta" % (cluster, dataset)
            fpath = os.path.join(weeder_fasta_dir, fname)
            fasta_files.append(fpath)

            with open(fpath, 'w') as outfile:
                for seq_name, seq in cluster_seqs.items():
                    outfile.write('>%s\n' % seq_name)
                    outfile.write('%s\n' % seq)
    return fasta_files


def find_motifs(fasta_files, tmpdir, freqfiles):
    # Setup for multiprocessing
    # Run this using all cores available
    cpus = cpu_count()
    print('Starting Weeder runs (%d CPUs available)...' % cpus, file=sys.stderr)
    pool = Pool(processes=cpus)
    pssms_list = pool.map(run_weeder, [(f, tmpdir, freqfiles) for f in fasta_files])
    print('Done with Weeder runs.', file=sys.stderr)

    # Compare to miRDB using my program
    final_pssms = [pssm for pssms in pssms_list for pssm in pssms]
    return [{'name': p.name,
             'sites': p.sites,
             'evalue': p.evalue,
             'genes': p.genes,
             'matrix': p.matrix}
            for p in final_pssms]


db = None
dataset = None
dataset_genes = None
total_targets = None
mirna_target_dict = None

def run_target_prediction_dbs(refSeq2entrez, exp_dir,
                              outdir, tmpdir, pred_db_dir, use_entrez):
    global db, dataset, dataset_genes, total_targets, mirna_target_dict

    mgr = Manager()
    dbs = [d for d in os.listdir(pred_db_dir)
           if os.path.isdir(os.path.join(pred_db_dir, d))]
    # Now do PITA and TargetScan - iterate through both platforms
    for db in dbs:
        print("checking against prediction database: '%s'" % db, file=sys.stderr)
        dboutdir = os.path.join(tmpdir, 'miRNA_%s' % db)
        if not os.path.exists(dboutdir):
            os.mkdir(dboutdir)


        # Get ready for multiprocessor goodness
        cpus = cpu_count()

        # Load up db of miRNA ids
        ls2 = [x for x in os.listdir(os.path.join(pred_db_dir, db))
               if x.endswith('.csv')]

        # Load the predicted target genes for each miRNA from the files
        tmp_dict = {}
        for f in ls2:
            miRNA = f.rstrip('.csv')
            with open(os.path.join(pred_db_dir, db, f), 'r') as inFile:
                tmp_dict[miRNA.lower()] = [int(line.strip()) for line in inFile.readlines()
                                          if line.strip()]
        mirna_target_dict = mgr.dict(tmp_dict)

        # Total background
        with open(os.path.join(pred_db_dir, db, db + '_ids_entrez.bkg'), 'r') as inFile:
            target_list = [int(x) for x in inFile.readlines() if x]
            tmp1 = target_list
            total_targets = mgr.list(tmp1)

        # For each cluster file in expfiles from Goodarzi et al.
        files = os.listdir(exp_dir)
        for f in files:
            # 3. Read in cluster file and convert to entrez ids
            print("File '%s'" % f)
            with open(os.path.join(exp_dir, f), 'r') as inFile:
                dataset = f.strip().split('.')[0]
                print("Data set: '%s'..." % dataset, file=sys.stderr)
                inFile.readline()
                lines = inFile.readlines()
                clusters = defaultdict(list)
                genes = []
                for line in lines:
                    gene, group = line.strip().split('\t')
                    group = int(group)
                    if use_entrez:
                        entrez = int(gene)
                    else:
                        if gene in refSeq2entrez:
                            entrez = refSeq2entrez[gene]
                        else:
                            entrez = None

                    if entrez in target_list:
                        genes.append(entrez)
                        clusters[group].append(entrez)

            dataset_genes = mgr.list(genes)

            # Iterate through clusters and compute p-value for each miRNA
            # Run this using all cores available
            pool = Pool(processes=cpus)
            params = [(cluster, cluster_genes, dataset, dboutdir)
                      for cluster, cluster_genes in clusters.items()]
            pool.map(cluster_hypergeo, params)
            print('Done.', file=sys.stderr)

        # 1. Get a list of all resulting overlap files in miRNA directory
        overlapFiles = os.listdir(dboutdir)

        # 2. Read them all in and grab the top hits
        with open(os.path.join(outdir, 'mergedResults_%s.csv' % db), 'w') as outFile:
            outFile.write('Dataset,Cluster,miRNA,q,m,n,k,p.value')
            enrichment = []
            for overlapFile in overlapFiles:
                overlap_path = os.path.join(dboutdir, overlapFile)
                with open(overlap_path, 'r') as inFile:
                    inFile.readline() # Get rid of header
                    lines = [line.strip().split(',') for line in inFile.readlines()]
                    miRNAs = [line[0].lstrip(db+'_') for line in lines]
                    intSect = [line[1] for line in lines]
                    miRNAPred = [line[2] for line in lines]
                    allNum = [line[3] for line in lines]
                    clustGenes = [line[4] for line in lines]
                    pVals = [float(line[5]) for line in lines]

                min1 = float(1)
                curMiRNA = []
                daRest = []
                for i in range(len(miRNAs)):
                    if pVals[i] < min1 and int(intSect[i])>=1:
                        min1 = pVals[i]
                        tmpMiRNA = miRNAs[i].lower()
                        if tmpMiRNA[-3:]=='-5p':
                            tmpMiRNA = tmpMiRNA[:-3]
                        curMiRNA = [tmpMiRNA]
                        daRest = [intSect[i], miRNAPred[i], allNum[i], clustGenes[i]]
                    elif pVals[i]==min1 and int(intSect[i])>=1:
                        tmpMiRNA = miRNAs[i].lower()
                        if tmpMiRNA[-3:]=='-5p':
                            tmpMiRNA = tmpMiRNA[:-3]
                        curMiRNA.append(tmpMiRNA)
                comps = overlapFile.rstrip('.csv').split('_')
                dataset = '_'.join(comps[:-1])
                cluster = comps[-1]
                outFile.write('\n' + dataset + ',' + cluster + ',' + ' '.join(curMiRNA) +
                              ',' + ','.join(daRest) + ',' + str(min1))
                enrichment.append({'dataset':dataset,
                                   'cluster':cluster,
                                   'miRNA':curMiRNA,
                                   'q' : daRest[0],
                                   'm' : daRest[1],
                                   'n' : daRest[2],
                                   'k' : daRest[3],
                                   'pValue' : min1,
                                   'percTargets' : float(daRest[0]) / float(daRest[3]),
                                   'significant' : False})

        # Filter using benjamini-hochberg FDR <= 0.001, >=10% target genes in cluster
        bhDict = {}
        for clust in range(len(enrichment)):
            bhDict[enrichment[clust]['dataset']+'_'+enrichment[clust]['cluster']] = enrichment[clust]['pValue']
        significant = benjamini_hochberg(bhDict, tests=len(clusters), alpha=0.001)
        # Do filtering
        filtered = []
        for clust in range(len(enrichment)):
            if (enrichment[clust]['dataset']+'_'+enrichment[clust]['cluster'] in significant) and (float(enrichment[clust]['q'])/float(enrichment[clust]['k']) >= 0.1):
                enrichment[clust]['significant'] = True
                filtered.append(enrichment[clust])

        # Write out filtered results
        fpath = os.path.join(outdir, 'filtered_%s.csv' % db)
        with open(fpath, 'w') as outFile:
            outFile.write('Dataset,Signature,miRNA,Percent.Targets')
            tot = 0
            for clust in range(len(filtered)):
                outFile.write('\n'+filtered[clust]['dataset']+','+filtered[clust]['cluster']+','+miRNA+','+str(float(enrichment[clust]['q'])/float(enrichment[clust]['k'])))


def write_combined_report(mirv_score_path, mirbase_path, outdir):
    mirna_ids = make_mirna_dict(mirbase_path)
    # Get miRvestigator results
    print("Retrieving miRvestigator results...", file=sys.stderr, end="", flush=True)
    miRNA_matches = {}
    with open(mirv_score_path,'r') as inFile:
        inFile.readline() # get rid of header
        lines = [i.strip().split(',') for i in inFile.readlines()]
        for line in lines:
            if not line[1]=='NA':
                miRNA_mature_seq_ids = []
                for i in line[1].split('_'):
                    miRNA_mature_seq_ids += mirnas_for(i.lower(), mirna_ids)
                cluster_name = [i for i in line[0].split('_')]
                cluster_name = cluster_name[1]+'_'+cluster_name[2]+'_'+cluster_name[3]+'_'+cluster_name[0]
                miRNA_matches[cluster_name] = {'miRNA':line[1],'model':line[2],'mature_seq_ids':miRNA_mature_seq_ids}

    print('Done.', file=sys.stderr)

    # Get PITA results
    print("Retrieving PITA results...", file=sys.stderr, end="", flush=True)
    with open(os.path.join(outdir, 'mergedResults_PITA.csv'), 'r') as inFile:
        inFile.readline() # get rid of header
        lines = [i.strip().split(',') for i in inFile.readlines()]

    for line in lines:
        if not line[2]=='':
            miRNA_mature_seq_ids = []
            mirs = [i.lower().strip('pita_') for i in line[2].split(' ')]
            for i in mirs:
                miRNA_mature_seq_ids += mirnas_for(i, mirna_ids)
            if not line[0]+'_'+line[1] in miRNA_matches:
                miRNA_matches[line[0]+'_'+line[1]] = {'pita_miRNA':' '.join(mirs),'pita_perc_targets':str(float(line[3])/float(line[6])),'pita_pValue':line[7],'pita_mature_seq_ids':miRNA_mature_seq_ids}
            else:
                miRNA_matches[line[0]+'_'+line[1]]['pita_miRNA'] = ' '.join(mirs)
                miRNA_matches[line[0]+'_'+line[1]]['pita_perc_targets'] = str(float(line[3])/float(line[6]))
                miRNA_matches[line[0]+'_'+line[1]]['pita_pValue'] = line[7]
                miRNA_matches[line[0]+'_'+line[1]]['pita_mature_seq_ids'] = miRNA_mature_seq_ids
    print('Done.', file=sys.stderr)

    # Get TargetScan results
    print("Retrieving Targetscan results...", file=sys.stderr, end="", flush=True)
    with open(os.path.join(outdir, 'mergedResults_TargetScan.csv'),'r') as inFile:
        inFile.readline() # get rid of header
        lines = [i.strip().split(',') for i in inFile.readlines()]

    for line in lines:
        if not line[2]=='':
            miRNA_mature_seq_ids = []
            mirs = [i.lower().strip('scan_') for i in line[2].split(' ')]
            for i in mirs:
                miRNA_mature_seq_ids += mirnas_for(i.lower().strip('targetscan_'), mirna_ids)
            if not line[0]+'_'+line[1] in miRNA_matches:
                miRNA_matches[line[0]+'_'+line[1]] = {'ts_miRNA':' '.join(mirs),'ts_perc_targets':str(float(line[3])/float(line[6])),'ts_pValue':line[7],'ts_mature_seq_ids':miRNA_mature_seq_ids}
            else:
                miRNA_matches[line[0]+'_'+line[1]]['ts_miRNA'] = ' '.join(mirs)
                miRNA_matches[line[0]+'_'+line[1]]['ts_perc_targets'] = str(float(line[3])/float(line[6]))
                miRNA_matches[line[0]+'_'+line[1]]['ts_pValue'] = line[7]
                miRNA_matches[line[0]+'_'+line[1]]['ts_mature_seq_ids'] = miRNA_mature_seq_ids
    print('Done.', file=sys.stderr)

    # Big list of all miRNAs for all clusters
    with open(os.path.join(outdir, 'combinedResults.csv'), 'w') as outFile:
        outFile.write('Dataset,signature,miRvestigator.miRNA,miRvestigator.model,miRvestigator.mature_seq_ids,PITA.miRNA,PITA.percent_targets,PITA.P_Value,PITA.mature_seq_ids,TargetScan.miRNA,TargetScan.percent_targets,TargetScan.P_Value,TargetScan.mature_seq_ids')
        for i in miRNA_matches:
            comps = i.split('_')
            dataset = '_'.join(comps[:-1])
            cluster = comps[-1]
            writeMe = '\n%s,%s' % (dataset, cluster)
            if 'miRNA' in miRNA_matches[i]:
                writeMe += ',' + miRNA_matches[i]['miRNA'] + ','+miRNA_matches[i]['model']+',' + ' '.join(miRNA_matches[i]['mature_seq_ids'])
            else:
                writeMe += ',NA,NA,NA'
            if 'pita_miRNA' in miRNA_matches[i]:
                writeMe += ',' + miRNA_matches[i]['pita_miRNA'] + ',' + miRNA_matches[i]['pita_perc_targets'] + ',' + miRNA_matches[i]['pita_pValue'] + ',' + ' '.join(miRNA_matches[i]['pita_mature_seq_ids'])
            else:
                writeMe += ',NA,NA,NA,NA'
            if 'ts_miRNA' in miRNA_matches[i]:
                writeMe += ',' + miRNA_matches[i]['ts_miRNA'] + ',' + miRNA_matches[i]['ts_perc_targets'] + ',' + miRNA_matches[i]['ts_pValue'] + ',' + ' '.join(miRNA_matches[i]['ts_mature_seq_ids'])
            else:
                writeMe += ',NA,NA,NA,NA'
            outFile.write(writeMe)
